#include <stdio.h>
#include "othello.h"

#define REST_FINAL 11

static void initialize(void);
static int aji_think(char* board);
static int aji_final(char* board);

typedef struct{
  int hand, point;
}HRESULT;

/***
  Main
****/

static int posp[2][16] = {
  { 001, 010, 021, 012, 030, 003, 023, 032,
      022, 013, 031, 011, 020, 002, 000, 033,},
  { 010, 001, 022, 003, 030, 032, 023,
      012, 021, 031, 013, 011, 002, 020, 000, 033,},
};

static int nSpace;
static int spaces[64];

int aji2(char* board, int status)
{
  int* p;
  int i;
  
  if(initp(status))
    initialize();
  if(passp(status) || initp(status)) return 0;
  
  for(nSpace = 0, i = 0; i < 64; i++)
    if(board[i] == SPACE) nSpace++;
  
  p = (nSpace > 30) ? posp[0] : posp[1];
  for(nSpace = 0, i = 0; i < 16; i++){
    if(board[p[i] ^ 000] == SPACE) spaces[nSpace++] = p[i] ^ 000;
    if(board[p[i] ^ 007] == SPACE) spaces[nSpace++] = p[i] ^ 007;
    if(board[p[i] ^ 070] == SPACE) spaces[nSpace++] = p[i] ^ 070;
    if(board[p[i] ^ 077] == SPACE) spaces[nSpace++] = p[i] ^ 077;
  }
  
  if(nSpace <= REST_FINAL)
    return aji_final(board);
  return aji_think(board);
}

/***
  util
***/

static int dir_step[4] = {1, 8, 9, 7};
static int check_lim[64][8]; /* [pos][dir] */

static void make_check_lim(void)
{
  int i, j;
  
  for(i = 0; i < 64; i += 8)
    for(j = i; j < i + 8; j++)
      check_lim[j][0] = i, check_lim[j][4] = i + 8;
  for(i = 0; i < 8; i++)
    for(j = i; j < 64; j += 8)
      check_lim[j][1] = 0, check_lim[j][5] = 64;
  for(i = 0; i < 8; i++)
    for(j = i; j < ((8 - i) << 3); j += 9)
      check_lim[j][2] = 0, check_lim[j][6] = ((8 - i) << 3);
  for(i = 8; i < 64; i += 8)
    for(j = i; j < 64; j += 9)
      check_lim[j][2] = i, check_lim[j][6] = 64;
  for(i = 0; i < 8; i++)
    for(j = i; j < (i << 3) + 1; j += 7)
      check_lim[j][3] = 0, check_lim[j][7] = (i << 3) + 1;
  for(i = 15; i < 64; i += 8)
    for(j = i; j < 64; j += 7)
      check_lim[j][3] = i, check_lim[j][7] = 64;
}

static int chk_rev(char* board, int pos, int col)
{
  int i, l, n, m, p, s, x;

  if(board[pos] != SPACE)
    return 0;
  x = (col == MY_COLOR) ? OP_COLOR : MY_COLOR;
  n = 0;
  for(i = 0; i < 4; i++){
    l = check_lim[pos][i], s = dir_step[i], m = 0;
    for(p = pos - s; p >= l; p -= s)
      if(board[p] == x){
	m++;
      }else{
	if(board[p] == col)
	  n += m;
	break;
      }
  }      
  for(; i < 8; i++){
    l = check_lim[pos][i], s = dir_step[i - 4], m = 0;
    for(p = pos + s; p < l; p += s)
      if(board[p] == x){
	m++;
      }else{
	if(board[p] == col)
	  n += m;
	break;
      }
  }
  return n;
}

static int reverse(char* board, int pos, int col)
{
  int i, l, n, m, p, s, x;

  if(board[pos] != SPACE)
    return 0;
  x = (col == MY_COLOR) ? OP_COLOR : MY_COLOR;
  n = 0;
  for(i = 0; i < 4; i++){
    l = check_lim[pos][i], s = dir_step[i], m = 0;
    for(p = pos - s; p >= l; p -= s)
      if(board[p] == x){
	m++;
      }else{
	if(board[p] == col){
	  n += m;
	  for(; m > 0; m--)
	    board[p += s] = col;
	}
	break;
      }
  }      
  for(; i < 8; i++){
    l = check_lim[pos][i], s = dir_step[i - 4], m = 0;
    for(p = pos + s; p < l; p += s)
      if(board[p] == x){
	m++;
      }else{
	if(board[p] == col){
	  n += m;
	  for(; m > 0; m--)
	    board[p -= s] = col;
	}
	break;
      }
  }
  if(n > 0)
    board[pos] = col;
  return n;
}


/***
  aji_think
***/

static int aji_think(char* board)
{
  int i;
  
  for(i = 0; i < nSpace; i++)
      if(chk_rev(board, spaces[i], MY_COLOR))
	return spaces[i];
  
  return -1;
}

/***
  aji_final
***/

#define FINAL_VAL_MAX 65
#define FINAL_VAL_MIN -65
static HRESULT aji_final_my(char* board, int min);
static int aji_final_pf;

static HRESULT aji_final_op(char* board, int max)
{
  HRESULT hr, hre;
  char bd[64];
  int i;
  
  hr.hand = -1, hr.point = FINAL_VAL_MIN;
  
  memcpy(bd, board, sizeof(bd));
  for(i = 0; i < nSpace; i++)
    if(board[spaces[i]] == SPACE)
      if(reverse(bd, spaces[i], OP_COLOR)){
	hre = aji_final_my(bd, hr.point);
	memcpy(bd, board, sizeof(bd));
	if(hr.hand == -1 || hr.point < hre.point){
	  hr.hand = spaces[i];
	  hr.point = hre.point;
	  if(max <= hr.point) break;
	}
      }
  if(hr.hand == -1)
    if(aji_final_pf){
      aji_final_pf = 0;
      hr.point = 0;
      for(i = 0; i< 64; i++)
	if(board[i] == MY_COLOR) hr.point++;
	else if(board[i] == OP_COLOR) hr.point--;
    }else{
      aji_final_pf = 1;
      hr = aji_final_my(board, FINAL_VAL_MIN);
    }
  else
    aji_final_pf = 0;
  
  return hr;
}

static HRESULT aji_final_my(char* board, int min)
{
  HRESULT hr, hre;
  char bd[64];
  int i;
  
  hr.hand = -1, hr.point = FINAL_VAL_MAX;
  
  memcpy(bd, board, sizeof(bd));
  for(i = 0; i < nSpace; i++)
    if(board[spaces[i]] == SPACE)
      if(reverse(bd, spaces[i], MY_COLOR)){
	hre = aji_final_op(bd, hr.point);
	memcpy(bd, board, sizeof(bd));
	if(hr.hand == -1 || hr.point > hre.point){
	  hr.hand = spaces[i];
	  hr.point = hre.point;
	  if(min >= hr.point)
	    break;
	}
      }
  if(hr.hand == -1)
    if(aji_final_pf){
      aji_final_pf = 0;
      hr.point = 0;
      for(i = 0; i< 64; i++)
	if(board[i] == MY_COLOR) hr.point++;
	else if(board[i] == OP_COLOR) hr.point--;
    }else{
      aji_final_pf = 1;
      hr = aji_final_op(board, FINAL_VAL_MAX);
    }
  else
    aji_final_pf = 0;
  
  return hr;
}

static int aji_final(char* board)
{
  HRESULT hr;

  aji_final_pf = 0;
  hr = aji_final_my(board, FINAL_VAL_MIN);

  return hr.hand;
}

/***
  initialize
***/

static void initialize(void)
{
  make_check_lim();
}
