/* othello.c Ver. 1.10 Written by Gel(Hideo KIHARA) */

#define __USE_FIXED_PROTOTYPES__

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>

#include "othello.h"
#include "othello2.h"

#define sec(t)  ((t)/1000000)
#define usec(t) ((t)%1000000)

static const char version[] = "X0.1.0";

const int direction_x[] = {-1,  0,  1, -1,    1, -1,  0,  1};   /* 0 1 2 */
const int direction_y[] = {-1, -1, -1,  0,    0,  1,  1,  1};   /* 3 @ 4 */
const int direction[] =   {-9, -8, -7, -1,    1,  7,  8,  9};   /* 5 6 7 */

static char m_board[64];
static int mode;

participant_t *think_fun[2], **fun_table;
int turn;
long think_t[2];
int entries;
static int last_hand;

int main(int argc, char **argv)
{
  int i, ret;
  int first, second;
  int **result;
  char c;
  
  X_initialize_application(&argc, argv);

  int flag = X_get_options();

  entries = 0;
  for(i = 0; e_list[i].func != NULL; i++, entries++);

  if(helpp(mode)){
    PrintHelp(entries);
    return 0;
  }

  X_create_gamewindow();
  X_create_selectwindow();

  XtAppMainLoop();
    
  return 0;
}

static void PlayGame(char *m_board, participant_t **think_fun,
		     const char *b_name, const char *w_name)
{
  int f, p, hands;
  char board[64];

  think_t[0] = think_t[1] = 0;
  last_hand = -1;

  CopyBoard(m_board, board, NORM);

  think_fun[0](board, INIT_AS_BLACK);
  think_fun[1](board, INIT_AS_WHITE);

  if(!silentp(mode) && !leaguep(mode)){
    PrintBoard(m_board, -1);
    WaitInput();
    printf("Player Black(%s)\'s turn.\n", b_name);
  }

  CopyBoard(m_board, board, NORM);
  f = 0;
  turn = 0;
  hands = 0;
  do{
    if(!isPass(board)){
      f = 0;
      
      SetTime();
      p = think_fun[turn](board, NORM);
      StopTime();

      last_hand = p;
      
      if((p < 0) || ( p > 63) || (Put(board, p, MY_COLOR, NOT_CHANGE) < 1)){
	printf("Player %s incorrect number : %d\n", (turn == 0)?"Black":"White", p);
	exit (-1);
      }

      hands++;
      Put(m_board, p, (turn == 0)?BLACK:WHITE, R_CHANGE);
      if(!silentp(mode) && !leaguep(mode))
	PrintBoard(m_board, p);
      RepairBoard(m_board);
      CopyBoard(m_board, board, (turn == 1)?REVERSE:NORM);

    }else{
      if(!silentp(mode) && !leaguep(mode)){
	printf("Player %s passed.\n", (turn == 0)?"Black":"White");
	PrintBoard(m_board, -1); /* ɤˤ֤ʤ */
      }
      think_fun[turn](board, PASS); /* pass */
      f++;
      last_hand = -1;
    }
    if(!silentp(mode) && !leaguep(mode)){
      WaitInput();
      printf("Player %s(%s)\'s turn.\n", (turn == 1)?"Black":"White", (turn ==1)?b_name:w_name);
    }
    turn = !turn;
    Reverse(board);
  }while((f != 2) && (hands < 60));
}

static void InitBoard(char *m_board)
{
  int i;

  /* Set Initial State */
  for(i = 0; i < 64; i++)
    m_board[i] = SPACE;

  m_board[27] = m_board[36] = WHITE; /*  */
  m_board[28] = m_board[35] = BLACK; /*  */
}

static void PrintBoard(char *board, int place)
{
  char *mark[] = {"", "", "", "", "", "", ""};
  int i, j, p;

  printf("\n");
  for(i = 0; i < 8; i++){
    printf("");
    for(j = 0; j < 8; j++){
      p = i * 8 + j;
      if(p == place)
	printf("%s",mark[(int)board[p] + 4]);
      else
	printf("%s",mark[(int)board[p] + 2]);
    }
    printf("\n");
    if(i < 7)
      printf("\n");
  }
  printf("\n");
}

static void RepairBoard(char *board)
{
  int i;

  for(i = 0; i < 64; i++)
    if(board[i] < 0)
      board[i] = -board[i];
}

static void WaitInput(void)
{
  printf("Hit return key.\n");
  getchar();
}

static int isPass(char *board)
{
  int i;

  for(i = 0; i < 64; i++)
    if(Put(board, i, MY_COLOR, NOT_CHANGE) > 0)
      return 0;
  return 1;
}

static void PrintHelp(int n)
{
  int i;

  printf("othello Ver. %s  Written by Gel(Hideo KIHARA)\n"
         "  Usage: othello [-s] num1 num2  Normal mode.\n"
	 "         othello [-s] -i         Interactive mode.\n"
	 "         othello [-s] -l         Full-League mode.\n"
	 "         othello [-s] -l num     Not-Full League(?) mode.\n"
	 , version);

  printf("\nParticipants:\n");

  for(i = 0; i < n; i++)
    printf("% 2d: %s\n", i, e_list[i].name);
}

static void PlayFullLeague(int **result, participant_t **fun_table)
{
  int i, j;
  int res_b, res_w;

  for(i = 0; i < (entries - 1); i++){
    for(j = 0; j < (entries - 1); j++){
      if(i == j)
	continue;
      think_fun[0] = fun_table[i];
      think_fun[1] = fun_table[j];
      
      printf("%-12s vs %-12s .. ", e_list[i + 1].name, e_list[j + 1].name);
      fflush(stdout);
      
      InitBoard(m_board);
      
      PlayGame(m_board, think_fun, e_list[i + 1].name, e_list[j + 1].name);
      
      CountDisks(m_board, &res_b, &res_w);
      
      printf("%2d-%2d\n", res_b, res_w);
      printf("%2ld.%06ldsec -- %2ld.%06ldsec\n", sec(think_t[0]), usec(think_t[0]),
	     sec(think_t[1]), usec(think_t[1]));
      
      if(!silentp(mode))
	PrintBoard(m_board, -1);
      
      result[j][i] = res_b - res_w;
    }
  }
}

static void PlaySemiLeague(int **result, participant_t **fun_table, int first)
{
  int i;
  int res_b, res_w;
  
  /* Black */
  think_fun[0] = fun_table[first - 1];
  
  for(i = 1; i < entries; i++){
    if(i == first)
      continue;
    think_fun[1] = fun_table[i - 1];
    
    InitBoard(m_board);
    
    printf("%-12s vs %-12s .. ", e_list[first].name, e_list[i].name);
    fflush(stdout);
    
    PlayGame(m_board, think_fun, e_list[first].name, e_list[i].name);
    
    CountDisks(m_board, &res_b, &res_w);
    
    printf("%2d-%2d\n", res_b, res_w);
    printf("%2ld.%06ldsec -- %2ld.%06ldsec\n", sec(think_t[0]), usec(think_t[0]),
	   sec(think_t[1]), usec(think_t[1]));
    
    if(!silentp(mode))
      PrintBoard(m_board, -1);
    
    result[0][(i < first)?(i - 1):(i - 2)] = res_b - res_w;
  }
  
  /* White */
  think_fun[1] = fun_table[first - 1];
  
  for(i = 1; i < entries; i++){
    if(i == first)
      continue;
    think_fun[0] = fun_table[i - 1];
    
    InitBoard(m_board);
    
    printf("%-12s vs %-12s .. ", e_list[i].name, e_list[first].name);
    fflush(stdout);
    
    PlayGame(m_board, think_fun, e_list[i].name, e_list[first].name);
    
    CountDisks(m_board, &res_b, &res_w);
    
    printf("%2d-%2d\n", res_b, res_w);
    printf("%2ld.%06ldsec -- %2ld.%06ldsec\n", sec(think_t[0]), usec(think_t[0]),
	   sec(think_t[1]), usec(think_t[1]));
    
    if(!silentp(mode))
      PrintBoard(m_board, -1);
    
    result[1][(i < first)?(i - 1):(i - 2)] = res_b - res_w;
  }
}

static int mat_alloc(int ***p, size_t size1, size_t size2)
{
  int i;

  *p = (int **)malloc(size1 * sizeof(int *));
  if(*p == NULL)
    return (-1);

  for(i = 0; i < size2; i++){
    (*p)[i] = (int *)malloc(size2 * sizeof(int));
    if((*p)[i] == NULL)
      return (-1);
  }

  return 0;
}

static void SelectPlayer(int *first, int *second)
{
  int i;
  char c;
  char buf[16];

  for(i = 0; i < entries; i++)
    printf("%d: %s\n", i, e_list[i].name);
  
  do{
    printf("Choose first player(Black):");
    fgets(buf, 16, stdin);
    if(feof(stdin))
      exit (-1);
    sscanf(buf, "%d", first);
  }while((*first < 0) || (*first >= entries));
  
  do{
    printf("Choose second player(White):");
    fgets(buf, 16, stdin);
    if(feof(stdin))
      exit (-1);
    sscanf(buf, "%d", second);
  }while((*second < 0) || (*second >= entries));
  
  do{
    printf("Display result only? (y/n)");
    fgets(buf, 16, stdin);
    if(feof(stdin))
      exit (-1);
    sscanf(buf, "%c", &c);
  }while((c != 'y') && (c != 'n') && (c != 'Y') && (c != 'N'));
  
  if((c == 'y') || (c == 'Y'))
    mode |= SILENT;
}

static void PlayNormGame(int first, int second)
{
  int res_b, res_w;

  /* Define Functions */
  think_fun[0] = e_list[first].func;
  think_fun[1] = e_list[second].func;
  
  InitBoard(m_board);
  
  PlayGame(m_board, think_fun, e_list[first].name, e_list[second].name);
  PrintBoard(m_board, -1);
  
  CountDisks(m_board, &res_b, &res_w);
  
  printf("Result: Black(%s)=%d [%2ld.%06ldsec]\n"
	 "        White(%s)=%d [%2ld.%06ldsec]\n",
	 e_list[first].name, res_b, sec(think_t[0]), usec(think_t[0]), 
	 e_list[second].name, res_w, sec(think_t[1]), usec(think_t[1]));
  
  if(res_b < res_w)
    printf("Player Black(%s) lose.\n", e_list[first].name);
  else if (res_b > res_w)
    printf("Player White(%s) lose.\n", e_list[second].name);
  else
    printf("Draw.(%s vs %s)\n", e_list[first].name, e_list[second].name);
  
}

int LastHand(void)
/* ľμ֤-1λϥѥ1ܤλ */
{
  return last_hand;
}
